/**config.c is part of JamMo.
License: GPLv2, read more from COPYING

This file is for clutter based gui and it handles config file reading
*/

#include "config.h"
#include <clutter/clutter.h>
#include <unistd.h> //for getcwd
#include <string.h>
#include "../chum/chum.h" //because of DATA_DIR

/**
 * The default ClutterScript config file.
 */
#define CONFIG_FILE_NAME g_strdup_printf("%s/config.json",DATA_DIR)

/*
Script is loaded everytime when new object is asked. 
If it is global we can't use same behaviour in two different buttons.
 And creating new object-actor doesn't work.
*/

#include <math.h>
gdouble  my_sine_wave (ClutterAlpha *alpha,gpointer user_data) {

return sin (clutter_timeline_get_progress (clutter_alpha_get_timeline (alpha)) * G_PI);
}

/**
 * config_return_script() returns a ClutterScript from the
 * parameter file given. Function returns NULL value if an
 * error occured.
 * 
 * @param		config_file		JSON-type file
 * @return		script			ClutterScript returned
 */
static ClutterScript* config_return_script (gchar* config_file)
{
	ClutterScript* script;
	char buffer[PATH_MAX];
	GError* error = NULL;
	
	script = clutter_script_new();
	
	if (!getcwd(buffer, PATH_MAX)) {
		fprintf(stderr, "Error: Current working directory path is too long.\n");
		return NULL;
	}
	
	if (!clutter_script_load_from_file(script, config_file, &error)) {
		fprintf(stderr, "return_script: Error: Could not load from file '%s' (%s).\n", config_file, error->message);
		g_error_free(error);
		return NULL;
	}
	
	clutter_script_connect_signals(script, NULL);
	
	return script;
}

ClutterScript* mainScript = NULL;
ClutterScript* sequencerScript = NULL;
ClutterScript* startmenuScript = NULL;
ClutterScript* gamesmenuScript = NULL;
ClutterScript* communitymenuScript = NULL;


/**
 * config_get_configured_actor() returns a ClutterActor according
 * to the parameter actor name given. The actor is returned from
 * the default config file, which is specified in the beginning
 * of this file. Function returns NULL value if an error occured.
 * Function uses config_return_script() function from this file.
 * 
 * @param		actor_name		actor wanted to be found
 * @return		actor			actor returned
 */
ClutterActor* config_get_configured_actor (const gchar *scriptname, gchar* actor_name)
{
	ClutterActor* actor; //this actor will be returned
	gchar *buffer;

	if (scriptname==NULL) //This means mainScript (general actors)
	{
		if (mainScript == NULL)  //If it isn't initialize, do it now.
		{
			buffer=g_strdup_printf("%s/config.json",DATA_DIR);
			if(!(mainScript = config_return_script(buffer)))
				{
				fprintf(stderr, "get_configured_actor: Error: Could not find config file.\n");
				g_free(buffer);
				return NULL;
				}
			g_free(buffer);
		}

		//Now mainScript is initialized
		if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(mainScript, actor_name)))) 
			{
			fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
			return NULL;
			}
		return actor;
	}
	
	//This is very same than previous.
	else if (strcmp(scriptname, "sequencer") == 0)
	{
		if (sequencerScript == NULL)
		{
			buffer=g_strdup_printf("%s/config.json",SEQUENCER_DIR);
			if(!(sequencerScript = config_return_script(buffer))) 
				{
				fprintf(stderr, "get_configured_actor: Error: Could not find config file.\n");
				g_free(buffer);
				return NULL;
			}
			g_free(buffer);
		}

		if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(sequencerScript, actor_name)))) {
			fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
			return NULL;
		}

		return actor;
	}
	else if (strcmp(scriptname, "startmenu") == 0)
		{
			if (startmenuScript == NULL)
			{
				buffer=g_strdup_printf("%s/config.json",STARTMENU_DIR);
				if(!(startmenuScript = config_return_script(buffer)))
					{
					fprintf(stderr, "get_configured_actor: Error: Could not find config file.\n");
					g_free(buffer);
					return NULL;
				}
				g_free(buffer);
			}

			if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(startmenuScript, actor_name)))) {
				fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
				return NULL;
			}

			return actor;
		}
	else if (strcmp(scriptname, "gamesmenu") == 0)
			{
				if (gamesmenuScript == NULL)
				{
					buffer=g_strdup_printf("%s/config.json",GAMESMENU_DIR);
					if(!(gamesmenuScript = config_return_script(buffer)))
						{
						fprintf(stderr, "get_configured_actor: Error: Could not find config file.\n");
						g_free(buffer);
						return NULL;
					}
					g_free(buffer);
				}

				if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(gamesmenuScript, actor_name)))) {
					fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
					return NULL;
				}

				return actor;
			}
	else if (strcmp(scriptname, "communitymenu") == 0)
			{
				if (communitymenuScript == NULL)
				{
					buffer=g_strdup_printf("%s/config.json",COMMUNITYMENU_DIR);
					if(!(communitymenuScript = config_return_script(buffer)))
						{
						fprintf(stderr, "get_configured_actor: Error: Could not find config file.\n");
						g_free(buffer);
						return NULL;
					}
					g_free(buffer);
				}

				if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(communitymenuScript, actor_name)))) {
					fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
					return NULL;
				}

				return actor;
			}
	else return NULL;
}


/*when theme starts, all objects are loaded to object_group (not with this function)
  when one object are respawned, this function are used to respawn.*/
ClutterActor* config_respawn_loop_actor (const gchar *theme_folder, gchar* actor_name) {
		gchar *buffer;
		ClutterActor* actor;
		//FIXME respawnig actor consumes large amounts of memory
		if (strcmp(theme_folder, "sequencer") == 0) { //needed for sequencer_loop
			buffer=g_strdup_printf("%s/config.json", SEQUENCER_DIR);
		} else {
			buffer=g_strdup_printf("%s/%s/config.json", "/tmp/jammo",theme_folder);
		}

		ClutterScript* temp_script;
		if(!(temp_script = config_return_script(buffer)))
			{
			fprintf(stderr, "get_objects_for_theme:Error: Could not find config file.\n");
			return NULL;
			}
		if(!(actor = CLUTTER_ACTOR(clutter_script_get_object(temp_script, actor_name)))){
			fprintf(stderr, "get_configured_actor: Error: Could not find %s from config file.\n", actor_name);
			return NULL;
		}
		g_free(buffer);
		g_object_unref(temp_script);
	return actor;
	
}



GList *config_get_objects_for_theme(const gchar *theme_folder){
  GList *actors = NULL;
  gchar* type="object";
  gchar *buffer;
  buffer=g_strdup_printf("%s/%s/config.json", "/tmp/jammo",theme_folder);
  ClutterScript* script;	
  if(!(script = config_return_script(buffer))) {
    fprintf(stderr, "get_objects_for_theme:Error: Could not find config file.\n");
    return NULL;
  }

  GList *all= clutter_script_list_objects (script);
  GList *l;
  for (l = all; l; l = l->next){
      if (strncmp(clutter_get_script_id(l->data),type,strlen(type)) ==0){
          clutter_actor_set_name(l->data, g_strdup_printf("%s_(%s)",clutter_get_script_id(l->data),theme_folder));
	  actors = g_list_append(actors, l->data); 
	  //printf("%s\n",clutter_get_script_id(l->data) );
      }
  }

return actors;
}






/**
 This function counts how many actor in config-file starts with given string.
 example:
 "button-" how many buttons
 
FIX: only buttons are counted in this way, and they should not be either.
*/
gint count_actors_in_script (gchar* type){
  ClutterScript* script;	

  char *buffer;
  char working_directory[PATH_MAX];
  if (!getcwd(working_directory, PATH_MAX)) {
    fprintf(stderr, "Error: Current working directory path is too long.\n");
    return -1;
  }
  buffer=g_strdup_printf("%s",CONFIG_FILE_NAME);



  if(!(script = config_return_script(buffer))) {
    fprintf(stderr, "Error: Could not find config file.\n");
    return -1;
  }
	
  gint count =0;
  GList *actors= clutter_script_list_objects (script);

  GList *l;
  for (l = actors; l; l = l->next){
    //printf("%s\n",clutter_get_script_id(l->data) );
    if (strncmp(clutter_get_script_id(l->data),type,strlen(type)) ==0){
      //printf("  count this\n");
      count++;
    }
  }
  return count;
}
